<?php

if (!class_exists(DatumConnector_REST_Controller::class)) {
    require_once(ABSPATH . 'wp-admin/includes/media.php');
    require_once(ABSPATH . 'wp-admin/includes/file.php');
    require_once(ABSPATH . 'wp-admin/includes/image.php');

    class DatumConnector_REST_Controller extends \WP_REST_Controller {
        function __construct() {
            // NOTE: Namespaces with the 'wc-' prefix are authenticated using the WC REST token.
            $this->namespace = "wc-datum/v1";
        }

        function register_routes() {
            register_rest_route($this->namespace, "/taxonomies", [
                "callback" => [$this, "get_taxonomies"],
                "permission_callback" => function() {
                    return current_user_can("manage_woocommerce");
                }
            ]);

            register_rest_route($this->namespace, "/sync-products", [
                "callback" => [$this, "sync_products"],
                "permission_callback" => function() {
                    return current_user_can("manage_woocommerce");
                }
            ]);
        }

        function get_taxonomies() {
            $remove_taxonomies = [
                "product_type",
                "product_visibility",
                "product_cat",
                "product_tag",
                "product_shipping_class"
            ];

            $taxonomies = array_filter(
                get_object_taxonomies("product"),
                fn($tax) => !in_array($tax, $remove_taxonomies) && strpos($tax, "pa_") !== 0
            );

            $result = [];
            foreach ($taxonomies as $tax) {
                $tax = get_taxonomy($tax);
                $result[] = ["label" => $tax->label, "name" => $tax->name];
            }

            wp_send_json($result);
        }

        function sync_products(\WP_REST_Request $request) {
            /** @var array<string, mixed>[] */
            $data = $request->get_json_params();
            $created = [];
            $updated = [];

            // Handle creates

            $creates = array_filter($data, fn ($array) => !array_key_exists("id", $array));

            foreach ($creates as $data) {
                if ($data["type"] == "simple") {
                    $product_id = $this->create_simple_product($data);
                    $created[$data["record_matching_key"]] = $product_id;
                }
            }

            // Handle updates

            $updates = array_filter($data, fn ($array) => array_key_exists("id", $array));

            /** @var array<\WC_Product> */
            $products = wc_get_products(['include' => array_column($updates, "id")]);

            foreach ($products as $product) {
                $found = null;

                foreach ($updates as $update) {
                    if ($update["id"] === $product->get_id()) {
                        $found = $update;
                        break;
                    }
                }

                if (!$found) {
                    // Product not found in WP so what do we do?
                    // Probably a user setting:
                    //   Either create or skip and warn
                } else {
                    $product_id = $this->update_product($product, $found);
                    $updated[$found["record_matching_key"]] = $product_id;
                }
            }

            // Handle images

            // The following will allow image sideloading with urls that don't
            // end in a 'known extension' such as jpg.
            // $wp_media_id = media_sideload_image( $url . '#.jpg', 0, null, 'id' );

            $new_images = [];

            foreach ($data as $item) {
                $images = $item["images"];

                if (count($images) == 0) {
                    continue;
                }

                $image_ids = [];

                foreach ($images as $image) {
                    if (is_int($image)) {
                        $image_ids[] = $image;
                    } elseif (is_string($image) && strpos($image, "https") === 0) {
                        /** @var int|WP_Error */
                        $id = @media_sideload_image($image, 0, null, 'id');
                        $image_ids[] = $id;
                        $new_images[$image] = $id;
                    } else {
                        error_log("found image value that is invalid: " . var_export($image, true));
                    }
                }

                $product_id = $created[$item["record_matching_key"]] ?? ($updated[$item["record_matching_key"]] ?? null);
                if (!$product_id) {
                    continue;
                }

                $product = wc_get_product($product_id);
                $product->set_image_id(array_shift($image_ids));
                $product->set_gallery_image_ids($image_ids);
                $product->save();
            }

            wp_send_json(["created" => $created, "i" => $new_images]);
        }

        /**
         * @param \WC_Product $product
         * @param array<string, mixed> $data
         * @return int
         */
        private function create_simple_product($data) {
            $post_id = wp_insert_post([
                "post_author" => get_current_user_id(),
                "post_title" => $data["name"],
                "post_type" => "product",
            ]);

            /** @var WC_Product */
            $product = wc_get_product($post_id);
            return $this->update_product($product, $data);
        }

        /**
         * @param \WC_Product $product
         * @param array<string, mixed> $data
         * @return void
         */
        private function update_product($product, $data) {
            $product->set_status($data['status'] ?? 'publish');
            $product->set_sku($data['sku'] ?? '');

            // Set descriptions
            $product->set_description($data['description'] ?? '');
            $product->set_short_description($data['short_description'] ?? '');

            // Set pricing
            $product->set_regular_price($data['regular_price'] ?? 0);
            $product->set_sale_price($data['sale_price'] ?? '');

            // Set dimensions and weight
            $product->set_weight($data['weight'] ?? '');
            $product->set_length($data['length'] ?? '');
            $product->set_width($data['width'] ?? '');
            $product->set_height($data['height'] ?? '');

            // Set stock management
            $product->set_manage_stock($data['manage_stock'] ?? false);
            $product->set_stock_quantity($data['stock_quantity'] ?? 0);
            $product->set_stock_status($data['stock_status'] ?? 'instock');
            $product->set_sold_individually($data['sold_individually'] ?? false);
            $product->set_backorders($data['backorders'] ?? 'no');

            // Taxonomies
            $product->set_category_ids($data["categories"]);

            $tag_ids = $this->get_or_create_tags($data["tags"]);
            $product->set_tag_ids($tag_ids);

            foreach ($data["taxonomies"] as $tax => $term) {
                $this->set_custom_taxonomies($product, $tax, $term);
            }

            // Set attributes
            $attrs = [];
            foreach ($data["attributes"] as $attr_data) {
                $name = wc_attribute_taxonomy_name_by_id($attr_data["id"]);
                $attr = new WC_Product_Attribute();
                $attr->set_id($attr_data["id"]);
                $attr->set_options($attr_data["options"]);
                $attr->set_variation($attr_data["variation"]);
                $attr->set_name($name);
                $attr->set_visible(true);
                $attr->set_position(0);
                $attrs[] = $attr;
            }
            $product->set_attributes($attrs);

            return $product->save();
        }

        private function get_or_create_tags($tags) {
            $tag_ids = [];
            foreach ($tags as $tag_name) {
                $term = term_exists($tag_name, 'product_tag');
                if (!$term) {
                    $term = wp_insert_term($tag_name, 'product_tag');
                }
                $tag_ids[] = $term['term_id'];
            }
            return $tag_ids;
        }

        private function set_custom_taxonomies($product, $tax, $term_name) {
            $term = term_exists($term_name, $tax);
            if (!$term) {
                $term = wp_insert_term($term_name, $tax);
            }
            wp_set_object_terms($product->get_id(), (int) $term['term_id'], $tax);
        }
    }
}
